import collections.abc
import itertools
import typing
import inspect
import types

from ..definition import contracts, exceptions


def register_module(
    registry: contracts.ActorRegistry,
    module: types.ModuleType
) -> contracts.ActorRegistry:
    for e in inspect.getmembers(module):
        entity = e[1]
        if inspect.ismodule(entity) and module.__name__ in entity.__name__:
            register_module(registry, entity)
        elif inspect.isfunction(entity) and not entity.__name__.startswith('_'):
            registry.add(entity)
    return registry


class ActorRegistry:
    def __init__(self):
        self._actors = {}
        self._contract_actor_map = collections.defaultdict(list)

    def add(
        self,
        actor: contracts.ActorContract
    ):
        if actor_entry := self._actor_entry(actor):
            self._register_entry(actor_entry)

    def get(
        self,
        contract: type
    ) -> collections.abc.Generator[contracts.ActorRegistryEntry, None, None]:
        chain = contract.__mro__ if hasattr(contract, '__mro__') else (contract,)
        for contract in chain:
            for entry_id in self._contract_actor_map.get(contract) or ():
                yield self._actors.get(entry_id)

    def get_by_id(
        self,
        actor_id: int
    ) -> contracts.ActorRegistryEntry | None:
        return self._actors.get(actor_id)

    def __iter__(self):
        for entry in self._actors.values():
            yield entry

    def __add__(self, other: contracts.ActorRegistry):
        for entry in other:
            self._register_entry(entry)
        return self

    def _register_entry(
        self,
        entry: contracts.ActorRegistryEntry
    ):
        self._actors[entry.id] = entry
        if entry.semantic_kind == contracts.ActorSemanticKind.CONTEXT:
            self._contract_actor_map[entry.return_annotation].append(entry.id)
        else:
            for parameter in itertools.chain(entry.parameters.domain, entry.parameters.context or ()):
                self._contract_actor_map[parameter.contract].append(entry.id)

    def _actor_entry(
        self,
        actor: contracts.ActorContract
    ) -> contracts.ActorRegistryEntry | None:
        semantic_kind = self._derive_semantic_kind(actor=actor)
        if semantic_kind is None:
            return
        signature = inspect.signature(actor)
        parameters = self._derive_parameters(signature)
        return_annotation = signature.return_annotation
        return contracts.ActorRegistryEntry(
            id=id(actor),
            parameters=parameters,
            return_annotation=return_annotation,
            semantic_kind=semantic_kind,
            runtime=contracts.ActorRuntime(
                executable=self._derive_executable(actor),
                kind=self._derive_execution_kind(actor)
            )
        )

    @staticmethod
    def _derive_semantic_kind(
        actor: contracts.ActorContract,
    ) -> contracts.ActorSemanticKind | None:
        _map = {
            'io': contracts.ActorSemanticKind.IO,
            'reader': contracts.ActorSemanticKind.IO,
            'writer': contracts.ActorSemanticKind.IO,
            'auditor': contracts.ActorSemanticKind.AUDIT,
            'usecase': contracts.ActorSemanticKind.USECASE,
            'context': contracts.ActorSemanticKind.CONTEXT,
            'response': contracts.ActorSemanticKind.RESPONSE,
            'skip': None
        }
        if '__semantic__' not in actor.__dict__:
            for item in _map.keys():
                if actor.__name__.endswith(f'_{item}'):
                    actor.__dict__['__semantic__'] = item
                    break
            else:
                return
        return _map[actor.__dict__['__semantic__']]

    @staticmethod
    def _derive_execution_kind(
        actor: contracts.ActorContract
    ) -> contracts.ActorExecutionKind:
        execution_type = contracts.ActorExecutionKind.SYNC
        if inspect.iscoroutinefunction(actor):
            execution_type = contracts.ActorExecutionKind.ASYNC

        return execution_type

    @staticmethod
    def _derive_executable(
        actor: contracts.ActorContract
    ):
        if inspect.isfunction(actor):
            return actor
        elif hasattr(actor, '__call__'):
            return actor.__call__

    def _derive_parameters(
        self,
        signature: inspect.Signature
    ) -> contracts.ActorParameters:
        domain = []
        dependencies = []
        context = []
        for parameter in signature.parameters.values():
            actor_parameter = self._derive_parameter(parameter)
            if isinstance(actor_parameter, contracts.ActorDomainParameter):
                domain.append(actor_parameter)
            if isinstance(actor_parameter, contracts.ActorContextParameter):
                context.append(actor_parameter)
            elif isinstance(actor_parameter, contracts.ActorDependencyParameter):
                dependencies.append(actor_parameter)
        if not domain:
            raise exceptions.CannotDefineActorParameter(signature)
        return contracts.ActorParameters(
            domain=domain,
            context=context or None,
            dependencies=dependencies or None
        )

    @staticmethod
    def _derive_parameter(parameter: inspect.Parameter) -> contracts.ActorParameter:
        _types_chain = _evoke_types_chain(parameter.annotation)
        if contract := _search_contract(_types_chain, contracts.Task, contracts.Event):
            return contracts.ActorDomainParameter(
                name=parameter.name,
                contract=contract
            )
        elif contract := _search_contract(_types_chain, contracts.Context):
            return contracts.ActorContextParameter(
                name=parameter.name,
                contract=contract
            )
        else:
            return contracts.ActorDependencyParameter(
                name=parameter.name,
                contract=parameter.annotation,
                default=contracts.NoValueType if parameter.default is inspect.Parameter.empty else parameter.default
            )


def _evoke_types_chain(annotation):
    _origin = typing.get_origin(annotation)
    if not _origin:
        return contracts.NoValueType if annotation is inspect.Parameter.empty else annotation
    else:
        args = typing.get_args(annotation)
        chain = [_origin]
        if len(args) == 1:
            r = _evoke_types_chain(args[0])
            if isinstance(r, collections.abc.Sequence):
                chain.extend(r)
            else:
                chain.append(r)
        else:
            chain.append(tuple(_evoke_types_chain(arg) for arg in args))
    return chain


def _search_contract(
    haystack: collections.abc.Sequence | type,
    *needles: type
) -> typing.Any:
    haystack = haystack if isinstance(haystack, collections.abc.Sequence) else (haystack,)
    if collections.abc.Callable in haystack:
        return
    for element in haystack:
        if isinstance(element, collections.abc.Sequence):
            if result := _search_contract(element, *needles):
                return result
        else:
            if any(
                needle is element or
                (hasattr(element, '__mro__') and needle in element.__mro__)
                for needle in needles
            ):
                return element



import zorge
from ..definition import contracts
from .processing import CQProcessor


class Executor:
    def __init__(
        self,
        di_container: zorge.Container,
        actor_registry: contracts.ActorRegistry
    ):
        self._di_container = di_container
        self._actor_registry = actor_registry

    async def run(
        self,
        message: contracts.Task,
        execution_context: contracts.ExecutionContext | None = None
    ) -> list[contracts.Message]:
        resolver_context = (execution_context,) if execution_context else ()
        stream = []
        async with self._di_container.get_resolver(*resolver_context) as resolver:
            processor = CQProcessor(self._actor_registry, resolver)(message)
            async for message in processor:
                stream.append(message)
        return stream



import dataclasses
import enum
import typing
import uuid
import collections.abc
import datetime

from zodchy import codex
from zodchy.types import Empty as EmptyType

IdentifierType: typing.TypeAlias = uuid.UUID
NoValueType: typing.TypeAlias = EmptyType
PacketIdType: typing.TypeAlias = IdentifierType
FrameIdType: typing.TypeAlias = IdentifierType
ActorIdType: typing.TypeAlias = int

Message: typing.TypeAlias = codex.cqea.Message
Task: typing.TypeAlias = codex.cqea.Task
Query: typing.TypeAlias = codex.cqea.Query
Context: typing.TypeAlias = codex.cqea.Context
Command: typing.TypeAlias = codex.cqea.Command
Event: typing.TypeAlias = codex.cqea.Event
Error: typing.TypeAlias = codex.cqea.Error
IOE: typing.TypeAlias = codex.cqea.IOEvent
BDE: typing.TypeAlias = codex.cqea.BDEvent
ResponseEvent: typing.TypeAlias = codex.cqea.ResponseEvent
ReadEvent: typing.TypeAlias = codex.cqea.ReadEvent
WriteEvent: typing.TypeAlias = codex.cqea.WriteEvent
Frame: typing.TypeAlias = codex.cqea.Frame
DIContainer: typing.TypeAlias = codex.di.DIContainerContract
DIResolver: typing.TypeAlias = codex.di.DIResolverContract
IdentifiersFactory: typing.TypeAlias = codex.identity.IdentifiersFactory
ActorContract: typing.TypeAlias = codex.cqea.Actor
ExecutionContext: typing.TypeAlias = collections.abc.Mapping[str, typing.Any]


class ActorExecutionKind(enum.Enum):
    SYNC = enum.auto()
    ASYNC = enum.auto()


class ActorSemanticKind(enum.Enum):
    AUDIT = enum.auto()
    USECASE = enum.auto()
    IO = enum.auto()
    CONTEXT = enum.auto()
    RESPONSE = enum.auto()


@dataclasses.dataclass
class ActorParameter:
    name: str
    contract: typing.Any


@dataclasses.dataclass
class ActorDomainParameter(ActorParameter):
    contract: type[Message] | type[Query]


@dataclasses.dataclass
class ActorContextParameter(ActorParameter):
    contract: type[Context]


@dataclasses.dataclass
class ActorDependencyParameter(ActorParameter):
    default: typing.Any


@dataclasses.dataclass
class ActorParameters:
    domain: collections.abc.Sequence[ActorDomainParameter]
    context: collections.abc.Sequence[ActorContextParameter] | None = None
    dependencies: collections.abc.Sequence[ActorDependencyParameter] | None = None


@dataclasses.dataclass
class ActorRuntime:
    executable: collections.abc.Callable
    kind: ActorExecutionKind


@dataclasses.dataclass
class ActorRegistryEntry:
    id: ActorIdType
    semantic_kind: ActorSemanticKind
    parameters: ActorParameters
    return_annotation: typing.Any
    runtime: ActorRuntime


class ActorRegistry(typing.Protocol):
    def add(self, actor: ActorContract): ...

    def get(self, contract: type) -> collections.abc.Generator[ActorRegistryEntry, None, None]: ...

    def get_by_id(self, actor_id: int) -> ActorRegistryEntry | None: ...

    def __iter__(self) -> collections.abc.Generator[ActorRegistryEntry, None, None]: ...

    def __add__(self, other: typing.Self) -> typing.Self: ...



import abc
from . import contracts


class PanchoException(Exception):
    pass


class CannotRegisterActor(PanchoException):
    def __init__(self, actor: contracts.ActorContract):
        self._actor = actor
        super().__init__(self.message())

    @abc.abstractmethod
    def message(self):
        return f'Cannot register actor: {self._actor}'


class CannotDeriveActorPurpose(PanchoException):
    def __init__(self, actor: contracts.ActorContract):
        self._actor = actor
        super().__init__(self.message())

    def message(self):
        return f'Cannot derive actor type: {self._actor}'


class CannotDefineActorParameter(PanchoException):
    def __init__(self, actor: contracts.ActorContract):
        self._actor = actor
        super().__init__(self.message())

    def message(self):
        return f'Cannot define actor parameter: {self._actor}'


class CannotResolveActorParameter(PanchoException):
    def __init__(self, actor_id: contracts.ActorIdType, param_name: str):
        self._actor_id = actor_id
        self._param_name = param_name
        super().__init__(self.message())

    def message(self):
        return f'Cannot resolve actor parameter: {self._param_name} of {self._actor_id}'


class CannotProcessActorResult(PanchoException):
    def __init__(self, actor_id: contracts.ActorIdType):
        self._actor_id = actor_id
        super().__init__(self.message())

    def message(self):
        return f'Cannot process actor result: {self._actor_id}'


class ActorSemanticDefinitionFailed(PanchoException):
    def __init__(self, actor_name: str):
        self._name = actor_name
        super().__init__(self.message())

    def message(self):
        return f'Actor semantic definition failed: {self._name}'



import typing
import functools


def semantic(kind: typing.Literal['usecase', 'io', 'auditor', 'context', 'response', 'reader', 'writer']):
    def decorator(func):
        func.__dict__['__semantic__'] = kind

        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            return await func(*args, **kwargs)

        return wrapper

    return decorator


def skip(func):
    func.__dict__['__semantic__'] = 'skip'

    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        return await func(*args, **kwargs)

    return wrapper



import collections.abc
import dataclasses
import typing
import heapq
import itertools
from os import write

import zorge

from ..definition import contracts, exceptions


@dataclasses.dataclass
class Job:
    actor_entry: contracts.ActorRegistryEntry
    parameters: collections.abc.Mapping[str, contracts.Message]


class Loop:
    _SEMANTIC_PRIORITY = {
        contracts.ActorSemanticKind.CONTEXT: 0,
        contracts.ActorSemanticKind.AUDIT: 1,
        contracts.ActorSemanticKind.USECASE: 2,
        contracts.ActorSemanticKind.IO: 3,
        contracts.ActorSemanticKind.RESPONSE: 9
    }

    def __init__(
        self,
        actor_registry: contracts.ActorRegistry
    ):
        self._queue = []
        self._actor_registry = actor_registry
        self._stream = {}

    def register(self, message: contracts.Message):
        if (key := message.__class__.__name__) not in self._stream:
            self._stream[key] = message
            self._register_job(message)

    def _register_job(self, message: contracts.Message):
        is_context = isinstance(message, contracts.Context)
        for actor_entry in self._actor_registry.get(message.__class__):
            if is_context and actor_entry.semantic_kind == contracts.ActorSemanticKind.CONTEXT:
                continue
            for context_parameter in actor_entry.parameters.context or ():
                self._register_context_job(context_parameter.contract)
            self._enqueue_job(actor_entry)

    def _register_context_job(self, contract: type[contracts.Context]):
        if contract.__name__ in self._stream:
            return
        for actor_entry in self._actor_registry.get(contract):
            if actor_entry.semantic_kind == contracts.ActorSemanticKind.CONTEXT:
                self._enqueue_job(actor_entry)

    def _enqueue_job(self, actor_entry: contracts.ActorRegistryEntry):
        if (parameters := self._build_parameters(actor_entry)) is not None:
            heapq.heappush(
                self._queue,
                (
                    self._SEMANTIC_PRIORITY[actor_entry.semantic_kind],
                    Job(
                        actor_entry=actor_entry,
                        parameters=parameters
                    )
                )
            )

    def _build_parameters(
        self,
        actor_entry: contracts.ActorRegistryEntry,
    ):
        parameters = {}
        for p in itertools.chain(actor_entry.parameters.domain, actor_entry.parameters.context or ()):
            if p.contract.__name__ not in self._stream:
                return
            parameters[p.name] = self._stream[p.contract.__name__]
        return parameters

    async def __aiter__(self):
        while self._queue:
            yield heapq.heappop(self._queue)[1]


class CQProcessor:
    def __init__(
        self,
        actor_registry: contracts.ActorRegistry,
        di_resolver: contracts.DIResolver | None = None
    ):
        self._actor_registry = actor_registry
        self._di_resolver = di_resolver

    async def __call__(
        self,
        message: contracts.Message
    ) -> typing.AsyncGenerator[contracts.Message, None]:
        loop = Loop(self._actor_registry)
        loop.register(message)
        async for job in loop:
            for message in await self._run_job(job):
                yield message
                if isinstance(message, contracts.Error):
                    return
                loop.register(message)

    async def _run_job(self, job: Job):
        params = {
            **job.parameters,
            **await self._compile_dependency_parameters(job)
        }
        if job.actor_entry.runtime.kind == contracts.ActorExecutionKind.ASYNC:
            result = await job.actor_entry.runtime.executable(**params)
        else:
            result = job.actor_entry.runtime.executable(**params)
        if not isinstance(result, collections.abc.Iterable):
            if result is None:
                result = ()
            elif isinstance(result, contracts.Message):
                result = (result,)
            else:
                raise ValueError(
                    f'Unexpected result type for actor {job.actor_entry.runtime.executable.__name__}: {type(result)}'
                )
        return result

    async def _compile_dependency_parameters(self, job: Job):
        params = {}
        if job.actor_entry.parameters.dependencies:
            for dependency_parameter in job.actor_entry.parameters.dependencies:
                if self._di_resolver:
                    params[dependency_parameter.name] = await self._di_resolver.resolve(dependency_parameter.contract)
                else:
                    if dependency_parameter.default is contracts.NoValueType:
                        raise exceptions.CannotResolveActorParameter(
                            actor_id=job.actor_entry.id,
                            param_name=dependency_parameter.name
                        )
                    params[dependency_parameter.name] = dependency_parameter.default
        return params


